% textpr -- convert text file for printing.
%
%      output = textpr(input, ...)
%
% The 'textpr' command generates a PostScript file from a text file.
%
% Form feed characters in the input cause page breaks in the output.
% Multiple form feed characters produce empty pages.  Unix, DOS, or
% Mac end-of-line character sequences cause line breaks.  Empty lines
% at the end of a page and whitespace characters at the end of a line
% are discarded.  Tab characters are expanded to spaces.
%
% Paper format, paper margins, page orientation, page width (number of
% characters per line), page length (number of lines per page), line
% spacing, font size, and n-up printing can be customized.
%
% Argument INPUT is the input file name.
% The rest of the argument list are options.
%
% Return value OUTPUT is the PostScript file name.
%
% Options:
%
% -title STRING
%      The job title (a string).  Default is the base file name of the
%      input file.  If set to the empty string, no DSC title comment will
%      be added to the PostScript file.
%
% -media NAME
%      The document paper format (a string).  Valid values for NAME are
%      'A4', 'A3', or 'letter'.  The default paper format is A4.
%
% -A4
%      Select A4 paper format.
%
% -letter
%      Select letter paper format.
%
% -left-margin NUMBER
%      The width of the left margin in PostScript points.  The default is
%      54 points, i.e. 3/4 inch or approximately 19.05 millimeter.
%
% -top-margin NUMBER
%      The height of the top margin in PostScript points.  The default is
%      18 points, i.e. 1/4 inch or approximately 6.35 millimeter.
%
% -right-margin NUMBER
%      The width of the right margin in PostScript points.  The default is
%      18 points, i.e. 1/4 inch or approximately 6.35 millimeter.
%
% -bottom-margin NUMBER
%      The height of the bottom margin in PostScript points.  The default
%      is 18 points, i.e. 1/4 inch or approximately 6.35 millimeter.
%
% -portrait
%      Use the sheet as is.  This is the default.
%
% -landscape
%      Rotate the sheet 90 degree clockwise.  This option rotates the
%      margins, too.
%
% -rows NUMBER
%      The number of virtual page rows per physical page.  The default
%      number of rows is one.
%
% -row-gap NUMBER
%      The extra space inserted between two virtual page rows in
%      PostScript points.  The default row gap is 18 points, i.e.
%      1/4 inch or approximately 6.35 millimeter.
%
% -columns NUMBER
%      The number of virtual page columns per physical page.  The default
%      number of columns is one.
%
% -column-gap NUMBER
%      The extra space inserted between two virtual page columns in
%      PostScript points.  The default column gap is 18 points, i.e.
%      1/4 inch or approximately 6.35 millimeter.
%
% -separators
%      Draw separator lines between virtual pages.  The default is to omit
%      separator lines.
%
% -line-width NUMBER
%      The separator line width in PostScript points.  The default line
%      width is 0.24 points.
%
% -font-size NUMBER
%      The font size in PostScript points.  This is the maximum font size
%      used to render text.  The default maximum font size is 10 points.
%
%      Once the page size is known, 'textpr' scales the font size so that
%      the page contents fits into it.  If that font size is larger than
%      the maximum font size, the maximum font size will be used instead.
%
% -characters-per-line NUMBER
%      The page width in characters per line.  If NUMBER is zero, the page
%      width is equal to the longest line of the input file contents, i.e.
%      no line will be truncated.  The default page width is 80 characters
%      per line.
%
% -lines-per-page NUMBER
%      The page length in lines per page.  If NUMBER is zero, the page
%      length is equal to the longest page of the input file contents,
%      i.e. no implicit page breaks occurs.  The default page length is
%      66 lines per page.
%
% -line-spacing NUMBER
%      The ratio of the baseline skip to the font size.  The default line
%      spacing is 1.2, i.e. an extra vertical space of 0.2 times the font
%      size is inserted between two consecutive lines.
%
% -anchor WHERE
%      The alignment of the page contents bounding box in the page.
%      Argument WHERE must be one of 'north-west', 'north', 'north-east',
%      'west', 'center', 'east', 'south-west', 'south', or 'south-east'.
%      The default anchor position is 'center'.
%
%      For example, 'north-west' means that the upper left corner of the
%      page contents bounding box is at the upper left corner of the page.
%
% -tab-width NUMBER
%      The distance (number of characters) between two adjacent tab stop
%      positions.  The default tab width is 8 characters.
%
% -chop-mark SYMBOL
%      The symbol to be placed at the end of a truncated line.  Argument
%      SYMBOL must be one of 'none', 'arrow', 'double-arrow', 'return', or
%      'ellipsis'.  The default chop mark is 'arrow'.
%
% -pass-through
%      Do not convert the input file into a PostScript file.  Instead, use
%      the input file as is for further processing.
%
% -ps FILENAME
%      Write the generated PostScript code to the file FILENAME.  The
%      default is to create a temporary file.  This option has no effect
%      in conjunction with the -pass-through option.
%
% -pdf FILENAME
%      Convert the PostScript file to PDF (Portable Document Format) and
%      save the PDF file in the file FILENAME.
%
% -ghostscript FILENAME
%      Utilize Ghostscript executable FILENAME for PostScript conversion.
%      Default is to use the Ghostscript executable shipped with Matlab.
%
% -version
%      Return version information and exit.
%
% The 'textpr' command uses Ghostscript to create non-PostScript files.
% In Matlab version 7.3 and newer, Ghostscript is packed into a private
% MEX-file which can't be called from a non-toolbox command.  For these
% versions of Matlab, 'textpr' depends on a separate Ghostscript
% installation.  If no Ghostscript executable is specified via the
% '-ghostscript' option, 'textpr' queries the value of the environment
% variable 'GS_EXE', then falls back to the 'gs' program.
%
% Examples:
%
% Convert the text file 'report.txt' into a 2-up PostScript file:
%
%      textpr('-landscape', '-columns', 2, ...
%             '-ps', 'report.ps', 'report.txt');
%
% Convert the text file 'report.txt' into a PDF file without saving the
% intermediate PostScript file:
%
%      delete(textpr('-pdf', 'report.pdf', 'report.txt'));
%
% Convert the paginated text file 'report.txt' into a 4-up PostScript file
% with separator lines so that each text file page fits into one virtual
% page:
%
%      textpr -rows=2 -columns=2 -separators ...
%             -characters-per-line=0 -lines-per-page=0 ...
%             -ps report.ps report.txt

%% textpr.m --- the preceding comment is the documentation string.

% Copyright (C) 2006 Ralph Schleicher

% Time-stamp: <2014-07-08 05:08:17 CEST>

% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions
% are met:
%
%    * Redistributions of source code must retain the above copyright
%      notice, this list of conditions and the following disclaimer.
%
%    * Redistributions in binary form must reproduce the above copyright
%      notice, this list of conditions and the following disclaimer in
%      the documentation and/or other materials provided with the
%      distribution.
%
%    * Neither the name of the copyright holder nor the names of its
%      contributors may be used to endorse or promote products derived
%      from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
% "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
% LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
% FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
% COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
% INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
% BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
% LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
% CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
% LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
% ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.

%% Code:

% Program entry point.
function output = textpr(varargin)

  % Initialize return value.
  output = sprintf(['textpr 1.2\n', ...
                    '\n', ...
                    'Copyright (C) 2006 Ralph Schleicher\n', ...
                    '\n', ...
                    'This program is free software and supplied as it is.  There is NO\n', ...
                    'WARRANTY; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR\n', ...
                    'PURPOSE."']);

  % Initialize options.
  opt = getopt(28);

  for ind = 1:numel(opt)
    switch ind
     case 1
      title = [];
      opt(ind).name = '-title';
      opt(ind).type = 'char';
      opt(ind).value = title;
      opt(ind).assign = 1;
     case 2
      media = 'A4';
      opt(ind).name = '-media';
      opt(ind).type = 'char';
      opt(ind).choices = {'A4', 'letter', 'A3'};
      opt(ind).value = media;
      opt(ind).assign = 1;
     case 3
      opt(ind).name = '-A4';
      opt(ind).type = 'alias';
      opt(ind).index = 2;
      opt(ind).value = 'A4';
     case 4
      opt(ind).name = '-letter';
      opt(ind).type = 'alias';
      opt(ind).index = 2;
      opt(ind).value = 'letter';
     case 5
      left_margin = 54;
      opt(ind).name = '-left-margin';
      opt(ind).type = 'non_negative_number';
      opt(ind).value = left_margin;
      opt(ind).assign = 1;
     case 6
      top_margin = 18;
      opt(ind).name = '-top-margin';
      opt(ind).type = 'non_negative_number';
      opt(ind).value = top_margin;
      opt(ind).assign = 1;
     case 7
      right_margin = 18;
      opt(ind).name = '-right-margin';
      opt(ind).type = 'non_negative_number';
      opt(ind).value = right_margin;
      opt(ind).assign = 1;
     case 8
      bottom_margin = 18;
      opt(ind).name = '-bottom-margin';
      opt(ind).type = 'non_negative_number';
      opt(ind).value = bottom_margin;
      opt(ind).assign = 1;
     case 9
      portrait = 1;
      opt(ind).name = '-portrait';
      opt(ind).type = 'flag';
      opt(ind).value = portrait;
      opt(ind).assign = 1;
     case 10
      opt(ind).name = '-landscape';
      opt(ind).type = 'alias';
      opt(ind).index = 9;
      opt(ind).value = 0;
     case 11
      rows = 1;
      opt(ind).name = '-rows';
      opt(ind).type = 'positive_integer';
      opt(ind).value = rows;
      opt(ind).assign = 1;
     case 12
      row_gap = 18;
      opt(ind).name = '-row-gap';
      opt(ind).type = 'non_negative_number';
      opt(ind).value = row_gap;
      opt(ind).assign = 1;
     case 13
      columns = 1;
      opt(ind).name = '-columns';
      opt(ind).type = 'positive_integer';
      opt(ind).value = columns;
      opt(ind).assign = 1;
     case 14
      column_gap = 18;
      opt(ind).name = '-column-gap';
      opt(ind).type = 'non_negative_number';
      opt(ind).value = column_gap;
      opt(ind).assign = 1;
     case 15
      separators = 0;
      opt(ind).name = '-separators';
      opt(ind).type = 'flag';
      opt(ind).value = separators;
      opt(ind).assign = 1;
     case 16
      line_width = 0.24;
      opt(ind).name = '-line-width';
      opt(ind).type = 'positive_number';
      opt(ind).value = line_width;
      opt(ind).assign = 1;
     case 17
      font_size = 10;
      opt(ind).name = '-font-size';
      opt(ind).type = 'positive_number';
      opt(ind).value = font_size;
      opt(ind).assign = 1;
     case 18
      characters_per_line = 80;
      opt(ind).name = '-characters-per-line';
      opt(ind).type = 'non_negative_integer';
      opt(ind).value = characters_per_line;
      opt(ind).assign = 1;
     case 19
      lines_per_page = 66;
      opt(ind).name = '-lines-per-page';
      opt(ind).type = 'non_negative_integer';
      opt(ind).value = lines_per_page;
      opt(ind).assign = 1;
     case 20
      line_spacing = 1.2;
      opt(ind).name = '-line-spacing';
      opt(ind).type = 'positive_number';
      opt(ind).value = line_spacing;
      opt(ind).assign = 1;
     case 21
      anchor = 'center';
      opt(ind).name = '-anchor';
      opt(ind).type = 'char';
      opt(ind).choices = {'north-west', 'north', 'north-east', 'west', 'center', 'east', 'south-west', 'south', 'south-east'};
      opt(ind).value = anchor;
      opt(ind).assign = 1;
     case 22
      tab_width = 8;
      opt(ind).name = '-tab-width';
      opt(ind).type = 'positive_integer';
      opt(ind).value = tab_width;
      opt(ind).assign = 1;
     case 23
      chop_mark = 'arrow';
      opt(ind).name = '-chop-mark';
      opt(ind).type = 'char';
      opt(ind).choices = {'none', 'arrow', 'double-arrow', 'return', 'ellipsis'};
      opt(ind).value = chop_mark;
      opt(ind).assign = 1;
     case 24
      pass_through = 0;
      opt(ind).name = '-pass-through';
      opt(ind).type = 'flag';
      opt(ind).value = pass_through;
      opt(ind).assign = 1;
     case 25
      ps = [];
      opt(ind).name = '-ps';
      opt(ind).type = 'char';
      opt(ind).value = ps;
      opt(ind).assign = 1;
     case 26
      pdf = [];
      opt(ind).name = '-pdf';
      opt(ind).type = 'char';
      opt(ind).value = pdf;
      opt(ind).assign = 1;
     case 27
      show_version = 0;
      opt(ind).name = '-version';
      opt(ind).type = 'flag';
      opt(ind).value = show_version;
      opt(ind).assign = 'show_version';
     case 28
      gs_bin = [];
      opt(ind).name = '-ghostscript';
      opt(ind).type = 'char';
      opt(ind).value = gs_bin;
      opt(ind).assign = 'gs_bin';
     otherwise
      error('Should not happen');
    end
  end

  % Process command line arguments.
  [opt, arg] = getopt(opt, varargin{:});

  if show_version
    return;
  end

  % Check number of arguments.
  error(nargchk(1, 1, numel(arg)));

  % Input file name.
  input = arg{1};

  % Create PostScript file.
  if pass_through
    output = input;
  else
    % Gather job information.
    if ~ ischar(title)
      [dir, name, ext] = fileparts(input);
      title = [name, ext];
    end

    % Convert file contents into a cell array of pages.
    % Each page is a cell array of strings.
    page = textpr_read(input, tab_width);
    pages = numel(page);

    % Determine the maximum number of characters per line
    % and the number of lines per page.
    characters = zeros(size(page));
    lines = zeros(size(page));

    for i = 1:pages
      if ~ isempty(page{i})
        tem = cellfun('size', page{i}, 2);
        characters(i) = max(tem);
        lines(i) = numel(page{i});
      end
    end

    % Paper dimensions.
    switch media
     case 'A4'
      paper_width = 595;
      paper_height = 842;
     case 'letter'
      paper_width = 612;
      paper_height = 792;
     case 'A3'
      paper_width = 842;
      paper_height = 1191;
    end

    % Axes dimensions.
    if portrait
      axes_width = paper_width - left_margin - right_margin;
      axes_height = paper_height - top_margin - bottom_margin;
    else
      axes_width = paper_height - top_margin - bottom_margin;
      axes_height = paper_width - left_margin - right_margin;
    end

    % Page dimensions.
    page_width = axes_width;
    if columns > 1
      page_width = (page_width - (columns - 1) * column_gap) / columns;
    end

    page_height = axes_height;
    if rows > 1
      page_height = (page_height - (rows - 1) * row_gap) / rows;
    end

    pages_per_page = rows * columns;
    if pages_per_page == 1
      separators = 0;
    end

    % Font size and derived quantities.
    if characters_per_line == 0
      characters_per_line = max(1, max(characters));
    end
    if lines_per_page == 0
      lines_per_page = max(1, max(lines));
    end

    font_size = min(font_size, ...
                    fix(min(page_width ...
                            / (characters_per_line * 0.6), ...
                            page_height ...
                            / (lines_per_page * line_spacing)) ...
                        * 2) / 2);

    char_width = 0.6 * font_size;

    % For Courier, font ascent is 0.811 times the font size
    % and font descent is -0.237 times the font size.
    top_skip = 0.85 * font_size;
    baseline_skip = line_spacing * font_size;

    % Bounding box dimensions.
    box_width = characters_per_line * char_width;
    box_height = lines_per_page * baseline_skip;

    % Bounding box alignment.
    switch anchor
     case 'north-west'
      horiz_align = 'left';
      vert_align = 'top';
     case 'north'
      horiz_align = 'center';
      vert_align = 'top';
     case 'north-east'
      horiz_align = 'right';
      vert_align = 'top';
     case 'west'
      horiz_align = 'left';
      vert_align = 'center';
     case 'center'
      horiz_align = 'center';
      vert_align = 'center';
     case 'east'
      horiz_align = 'right';
      vert_align = 'center';
     case 'south-west'
      horiz_align = 'left';
      vert_align = 'bottom';
     case 'south'
      horiz_align = 'center';
      vert_align = 'bottom';
     case 'south-east'
      horiz_align = 'right';
      vert_align = 'bottom';
     otherwise
      error('Should not happen');
    end

    switch horiz_align
     case 'left'
      horiz_offset = 0;
     case 'center'
      horiz_offset = (page_width - box_width) / 2;
     case 'right'
      horiz_offset = page_width - box_width;
     otherwise
      error('Should not happen');
    end

    switch vert_align
     case 'top'
      vert_offset = 0;
     case 'center'
      vert_offset = (page_height - box_height) / 2;
     case 'bottom'
      vert_offset = page_height - box_height;
     otherwise
      error('Should not happen');
    end

    % Absolute line length limit (i.e. string position to cut).
    chop_index = 1 + fix((page_width - horiz_offset) / char_width);

    % Emit PostScript code.
    if ischar(ps)
      output = ps;
    else
      output = [tempname, '.ps'];
    end

    h = fopen(output, 'w');
    if h == -1
      error(sprintf('Can not open file ''%s''', output));
    end

    % Prolog.
    if portrait
      ps_portrait = 'true';
      dsc_orientation = 'Portrait';
    else
      ps_portrait = 'false';
      dsc_orientation = 'Landscape';
    end
    if pages_per_page == 1
      dsc_page_order = 'Ascend';
    else
      dsc_page_order = 'Special';
    end

    switch chop_mark
     case 'none'
      ps_chop_mark = 'space';
     case 'arrow'
      ps_chop_mark = 'arrowdown';
     case 'double-arrow'
      ps_chop_mark = 'arrowdbldown';
     case 'return'
      ps_chop_mark = 'carriagereturn';
     otherwise
      ps_chop_mark = chop_mark;
    end

    fprintf(h, '%%!PS-Adobe-3.0\n');
    if ~ isempty(title)
      fprintf(h, '%%%%Title: %s\n', textpr_quote(title, 7));
    end
    fprintf(h, '%%%%Creator: %s.m\n', mfilename);
    fprintf(h, '%%%%DocumentData: Clean8Bit\n');
    fprintf(h, '%%%%DocumentMedia: %s %d %d 0 () ()\n', ...
            media, paper_width, paper_height);
    fprintf(h, '%%%%DocumentNeededResources: font Courier Symbol\n');
    fprintf(h, '%%%%Pages: (atend)\n');
    fprintf(h, '%%%%PageOrder: %s\n', dsc_page_order);
    fprintf(h, '%%%%Orientation: %s\n', dsc_orientation);
    fprintf(h, '%%%%EndComments\n');

    fprintf(h, '%%%%BeginDefaults\n');
    fprintf(h, '%%%%PageMedia: %s\n', media);
    fprintf(h, '%%%%PageOrientation: %s\n', dsc_orientation);
    fprintf(h, '%%%%EndDefaults\n');

    fprintf(h, '%%%%BeginProlog\n');
    fprintf(h, '/textpr-paper-width %G def\n', paper_width);
    fprintf(h, '/textpr-paper-height %G def\n', paper_height);
    fprintf(h, '/textpr-left-margin %G def\n', left_margin);
    fprintf(h, '/textpr-top-margin %G def\n', top_margin);
    fprintf(h, '/textpr-right-margin %G def\n', right_margin);
    fprintf(h, '/textpr-bottom-margin %G def\n', bottom_margin);
    fprintf(h, '/textpr-portrait %s def\n', ps_portrait);
    fprintf(h, '/textpr-axes-width %d def\n', axes_width);
    fprintf(h, '/textpr-axes-height %d def\n', axes_height);
    fprintf(h, '/textpr-rows %d def\n', rows);
    fprintf(h, '/textpr-row-gap %G def\n', row_gap);
    fprintf(h, '/textpr-columns %d def\n', columns);
    fprintf(h, '/textpr-column-gap %G def\n', column_gap);
    fprintf(h, '/textpr-page-width %G def\n', page_width);
    fprintf(h, '/textpr-page-height %G def\n', page_height);
    fprintf(h, '/textpr-font-size %G def\n', font_size);
    fprintf(h, '/textpr-chop-mark /%s def\n', ps_chop_mark);
    fprintf(h, '\n');
    fprintf(h, '%% textpr-coordinates\n');
    fprintf(h, '/textpr-coordinates {\n');
    fprintf(h, '  textpr-portrait {\n');
    fprintf(h, '    textpr-left-margin textpr-bottom-margin translate\n');
    fprintf(h, '  } {\n');
    fprintf(h, '    textpr-left-margin textpr-axes-height add textpr-bottom-margin translate\n');
    fprintf(h, '    90 rotate\n');
    fprintf(h, '  } ifelse\n');
    fprintf(h, '} bind def\n');
    fprintf(h, '\n');
    fprintf(h, '%% textpr-separators\n');
    fprintf(h, '/textpr-separators {\n');
    fprintf(h, '  gsave\n');
    fprintf(h, '  %G setlinewidth\n', line_width);
    fprintf(h, '  1 1 textpr-rows {\n');
    fprintf(h, '    /textpr-i exch def\n');
    fprintf(h, '    /textpr-y textpr-rows textpr-i sub textpr-page-height textpr-row-gap add mul def\n');
    fprintf(h, '    /textpr-x textpr-page-width textpr-column-gap 2 div add def\n');
    fprintf(h, '    2 1 textpr-columns {\n');
    fprintf(h, '      pop\n');
    fprintf(h, '      newpath textpr-x textpr-y moveto 0 textpr-page-height rlineto stroke\n');
    fprintf(h, '      /textpr-x textpr-x textpr-page-width textpr-column-gap add add def\n');
    fprintf(h, '    } for\n');
    fprintf(h, '  } for\n');
    fprintf(h, '  1 1 textpr-columns {\n');
    fprintf(h, '    /textpr-j exch def\n');
    fprintf(h, '    /textpr-x textpr-columns textpr-j sub textpr-page-width textpr-column-gap add mul def\n');
    fprintf(h, '    /textpr-y textpr-page-height textpr-row-gap 2 div add def\n');
    fprintf(h, '    2 1 textpr-rows {\n');
    fprintf(h, '      pop\n');
    fprintf(h, '      newpath textpr-x textpr-y moveto textpr-page-width 0 rlineto stroke\n');
    fprintf(h, '      /textpr-y textpr-y textpr-page-height textpr-row-gap add add def\n');
    fprintf(h, '    } for\n');
    fprintf(h, '  } for\n');
    fprintf(h, '  grestore\n');
    fprintf(h, '} bind def\n');
    fprintf(h, '\n');
    fprintf(h, '%% NEWFONT OLDFONT textpr-recode\n');
    fprintf(h, '/textpr-recode {\n');
    fprintf(h, '  findfont\n');
    fprintf(h, '  dup length dict begin\n');
    fprintf(h, '    {1 index /FID ne {def} {pop pop} ifelse} forall\n');
    fprintf(h, '    /Encoding ISOLatin1Encoding def\n');
    fprintf(h, '    currentdict end\n');
    fprintf(h, '  definefont pop\n');
    fprintf(h, '} bind def\n');
    fprintf(h, '\n');
    fprintf(h, '/Courier-ISO-8859-1 /Courier textpr-recode\n');
    fprintf(h, '\n');
    fprintf(h, '/textpr-regular /Courier-ISO-8859-1 findfont textpr-font-size scalefont def\n');
    fprintf(h, '/textpr-symbol /Symbol findfont textpr-font-size scalefont def\n');
    fprintf(h, '\n');
    fprintf(h, '/R {\n');
    fprintf(h, '  textpr-regular setfont\n');
    fprintf(h, '} bind def\n');
    fprintf(h, '\n');
    fprintf(h, '/M {\n');
    fprintf(h, '  moveto\n');
    fprintf(h, '} bind def\n');
    fprintf(h, '\n');
    fprintf(h, '/S {\n');
    fprintf(h, '  show\n');
    fprintf(h, '} bind def\n');
    fprintf(h, '\n');
    fprintf(h, '/C {\n');
    fprintf(h, '  currentfont\n');
    fprintf(h, '  textpr-symbol setfont\n');
    fprintf(h, '  textpr-chop-mark glyphshow\n');
    fprintf(h, '  setfont\n');
    fprintf(h, '} bind def\n');
    fprintf(h, '%%%%EndProlog\n');

    % Script.
    physical_page = 0;
    page_number = 1;

    while page_number <= pages
      % Begin a new physical page.
      physical_page = physical_page + 1;
      if pages_per_page == 1
        fprintf(h, '%%%%Page: %d %d\n', ...
                physical_page, physical_page);
      else
        last = physical_page * pages_per_page;
        first = last - pages_per_page + 1;
        fprintf(h, '%%%%Page: %d-%d %d\n', ...
                first, last, physical_page);
      end
      fprintf(h, '%%%%BeginPageSetup\n');
      fprintf(h, '/textpr-state save def\n');
      fprintf(h, '%%%%EndPageSetup\n');
      fprintf(h, 'textpr-coordinates\n');
      if separators
        fprintf(h, 'textpr-separators\n');
      end
      % Typeset virtual pages.
      for i = 1:rows
        for j = 1:columns
          if pages_per_page > 1
            fprintf(h, '%% Virtual page %d/%d.\n', ...
                    (i - 1) * columns + j, pages_per_page);
          end
          if page_number <= pages
            fprintf(h, 'R\n');
            x = ((j - 1) * (page_width + column_gap) ...
                 + horiz_offset);
            y = ((rows - i) * (page_height + row_gap) ...
                 + page_height - vert_offset - top_skip);
            for k = 1:min(lines(page_number), lines_per_page)
              str = page{page_number}{k};
              if numel(str) < chop_index
                fprintf(h, '%G %G M\n(%s) S\n', ...
                        x, y, textpr_quote(str, 8));
              else
                str(chop_index:end) = [];
                fprintf(h, '%G %G M\n(%s) S C\n', ...
                        x, y, textpr_quote(str, 8));
              end
              y = y - baseline_skip;
            end
            if lines(page_number) > lines_per_page
              % Put remaining lines on the next virtual page.
              page{page_number}(1:lines_per_page) = [];
              lines(page_number) = numel(page{page_number});
            else
              % Process next page.
              page_number = page_number + 1;
            end
          else
            fprintf(h, '%% Intentionally blank.\n');
          end
        end
      end
      fprintf(h, 'textpr-state restore\n');
      fprintf(h, 'showpage\n');
    end

    fprintf(h, '%%%%Trailer\n');
    fprintf(h, '%%%%Pages: %d\n', physical_page);
    fprintf(h, '%%%%EOF\n');

    % Close PostScript file.
    if fclose(h) == -1
      error(sprintf('Can not close file ''%s''', output));
    end
  end

  % Utilize Ghostscript for PostScript conversion.
  if ~ ischar(pdf)
    return;
  end

  arch = lower(computer);
  switch arch
   case 'pcwin'
    arch = 'win32';
    exeext = '.exe';
   otherwise
    exeext = '';
  end

  % Check whether or not this Matlab version has a bundled
  % Ghostscript executable.
  %
  % Empty means to utilize an external Ghostscript interpreter.
  gs_prefix = [];

  if ~ isoctave
    ver = sscanf(version, '%d.%d');

    if ver(1) < 7 || (ver(1) == 7 && ver(2) < 3)
      % Only use an external Ghostscript interpreter if explicitly
      % specified on the command line.
      if ~ ischar(gs_bin)
	gs_prefix = fullfile(matlabroot, 'sys', 'ghostscript');
	if ~ exist(gs_prefix, 'dir')
	  gs_prefix = fullfile(matlabroot, 'ghostscript');
	  if ~ exist(gs_prefix, 'dir')
	    error('Can not locate top-level Ghostscript directory');
	  end
	end

	gs_bin = fullfile(gs_prefix, 'bin', arch, ['gs', exeext]);
	if ~ exist(gs_bin, 'file')
	  gs_bin = fullfile(gs_prefix, 'bin', ['gs', exeext]);
	  if ~ exist(gs_bin, 'file')
	    error('Can not locate Ghostscript executable');
	  end
	end
      end
    end
  end

  if isempty(gs_prefix)
    % Look for an external Ghostscript interpreter.
    if isempty(gs_bin)
      gs_bin = getenv('GS_EXE');
      if isempty(gs_bin)
        gs_bin = 'gs';
      end
    end

    % Sanity check.  Presume that the bundled Ghostscript
    % interpreter works.
    [err, out] = system(sprintf('"%s" -version', gs_bin));
    if err
      error(sprintf('Can not invoke Ghostscript executable ''%s''', gs_bin));
    end
    if strfind(out, 'Ghostscript')
      0;
    else
      error(sprintf('Program ''%s'' does not smell like Ghostscript', gs_bin));
    end
  end

  % Ghostscript arguments.
  gs_output = pdf;
  gs_device = 'pdfwrite';

  switch media
   case 'A4'
    gs_media = 'a4';
   case 'A3'
    gs_media = 'a3';
   otherwise
    gs_media = media;
  end

  % Create response file.
  resp = tempname;

  h = fopen(resp, 'w');
  if h == -1
    error('Can not create response file');
  end

  if ~ isempty(gs_prefix)
    fprintf(h, '-I"%s"\n', fullfile(gs_prefix, 'ps_files'));
    fprintf(h, '-I"%s"\n', fullfile(gs_prefix, 'fonts'));
  end

  fprintf(h, '-q -dSAVER -dNOPAUSE -dBATCH\n');
  fprintf(h, '-sOutputFile="%s"\n', gs_output);
  fprintf(h, '-sPAPERSIZE=%s\n', gs_media);
  fprintf(h, '-sDEVICE=%s\n', gs_device);
  fprintf(h, '"%s"\n', output);

  if fclose(h) == -1
    error('Can not close response file');
  end

  % Invoke Ghostscript.
  [err, out] = system(sprintf('"%s" "@%s"', gs_bin, resp));

  % Clean up.
  delete(resp);

  if err
    err = sprintf('Can not create file ''%s''', pdf);
    out = textpr_trim(out);
    if ~ isempty(out)
      err = sprintf('%s.\n\n%s', err, out);
    end
    error(err);
  end

% Read a text file.
function page = textpr_read(file_name, tab_width)

  % Slurp in file contents.
  h = fopen(file_name, 'r');
  if h == -1
    error(sprintf('Can not open file ''%s''', file_name));
  end

  str = fread(h);
  str = char(str');

  if fclose(h) == -1
    error(sprintf('Can not close file ''%s''', file_name));
  end

  % Control characters.
  tem = sprintf('\t\n\f\r');

  ht = tem(1);
  nl = tem(2);
  ff = tem(3);
  cr = tem(4);

  % End-of-line conversion.
  ind = strfind(str, [cr, nl]);
  if ind
    str(ind) = [];
  end
  ind = strfind(str, cr);
  if ind
    str(ind) = nl;
  end

  % Break file contents into pages.
  ind = strfind(str, ff);
  pages = numel(ind) + 1;
  page = cell(pages, 1);
  start = 1;

  for i = 1:numel(ind)
    tem = str(start:ind(i));
    page{i} = textpr_split(tem, nl);
    start = ind(i) + 1;
  end

  if start < numel(str)
    tem = str(start:end);
    page{end} = textpr_split(tem, nl);
  end

  % Expand tab characters.  There are no trailing tab characters
  % at the end of a line.
  sp = repmat(' ', 1, tab_width);

  for i = 1:pages
    for j = 1:numel(page{i})
      str = page{i}{j};
      if isempty(str)
        continue;
      end
      while 1
        ind = find(str == ht);
        if isempty(ind)
          break;
        end
        if numel(ind) > 1
          ind = ind(1);
        end
        str(ind)= [];
        ind = ind - 1;
        len = tab_width - rem(ind, tab_width);
        str = [str(1:ind), sp(1:len), str((ind + 1):end)];
      end
      page{i}{j} = str;
    end
  end

% Convert a formatted string into a cell string.
function str = textpr_split(str, nl)

  str = textpr_trim(str);
  if isempty(str)
    str = [];
    return;
  end

  % Create a cell array of strings.
  str = strrep(str, '''', '''''');
  str = strrep(str, nl, '''; ''');
  str = eval(['{''', str, '''}']);

  for i = 1:numel(str)
    str{i} = textpr_trim(str{i});
  end

% Remove trailing whitespace characters.
function str = textpr_trim(str)

  ind = find(~ isspace(str));
  if isempty(ind)
    str = '';
  else
    str = str(1:ind(end));
  end

% Return a proper Postscript string.
function str = textpr_quote(str, bits)

  if isempty(str)
    return;
  end

  % Quote special characters.
  if any(str == '\')
    str = strrep(str, '\', '\\');
  end
  if any(str == '(')
    str = strrep(str, '(', '\(');
  end
  if any(str == ')')
    str = strrep(str, ')', '\)');
  end

  % Replace all non-ASCII characters with their octal code.
  switch bits
   case 7
    ch = [0:31, 127:255];
   case 8
    ch = [0:31, 127:159];
   otherwise
    error('Should not happen');
  end

  for i = ch
    if any(str == i)
      str = strrep(str, char(i), sprintf('\\%03o', i));
    end
  end

function p = isoctave

  p = exist('OCTAVE_VERSION', 'builtin');

% local variables:
% time-stamp-line-limit: 400
% time-stamp-format: "%:y-%02m-%02d %02H:%02M:%02S %Z"
% page-delimiter: "^function\\>.*\n?"
% end:

%% textpr.m ends here
